/*
 * Copyright 2012 Sebastian Mihai
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * audio.c
 *
 *  Created on: 2012-08-09
 *      Author: Sebastian Mihai
 */


#include <sys/stat.h>
#include <assert.h>
#include <screen/screen.h>
#include <bps/bps.h>
#include <unistd.h>
#include <stdio.h>
#include <mm/renderer.h>

#include "audio.h"

/*
 *
 * We keep the following streams active at all times (they are opened when the program starts
 * and are closed when it ends) :
 *
 * - background music
 * - touch 1 (this can load either horn, dog, or the standard "click" - played when the screen is touched)
 * - sheep baa 1
 * - sheep baa 2
 * - bear growl
 * - touch 2 (same as touch 1)
 *
 */

// This variable tells us that something bad has happened at some point in time
// In order to keep the game playing, all audio functionality is disabled if this becomes 1
int generalSoundError = 0;

mmr_connection_t *mmrConnection[STREAM_COUNT];
mmr_context_t *mmrContext[STREAM_COUNT];

char *mmrName[STREAM_COUNT];
const char *mmrContextName[STREAM_COUNT] = { "bgm", "touch1", "sheepbaa1", "sheepbaa2", "beargrowl", "touch2", "horn" };
char *mmrAudioIdentifier[STREAM_COUNT];
const char *mmrInputType = "track";
char *mmrAudioIdentifierConstant = "audio:default";

/*
 * Private helper function - get file paths for sound effects and audio tracks
 */
char* get_filename_for_sound( int which ) {

	switch( which ) {
	case BGM_SOUTHERN : return "/app/native/assets/music/southern.mp3";
	case SFX_TOUCH_CLICK : return "/app/native/assets/sounds/click.wav";
	case SFX_BAA_1 : return "/app/native/assets/sounds/baa1.wav";
	case SFX_BAA_2 : return "/app/native/assets/sounds/baa2.wav";
	case SFX_PLACE_DOG : return "/app/native/assets/sounds/bear.wav";
	case SFX_BEAR_HORN : return "/app/native/assets/sounds/horn.wav";
	case SFX_BEAR_GROWL : return "/app/native/assets/sounds/bear.wav";
	}

	return SFX_TOUCH_CLICK;
}

// =================== Exposed functions ======================

void stream_play( int which ) {
	if( generalSoundError ) return;

	if ( mmr_play( mmrContext[which] ) < 0 ) {
		generalSoundError = 1; return;
	}
}

void stream_close( int which ) {
	if( generalSoundError ) return;
	mmr_stop( mmrContext[which] );
	mmr_input_detach( mmrContext[which] );
	mmr_context_destroy( mmrContext[which] );
	mmr_disconnect( mmrConnection[which] );
}

void initialize_audio() {
	int i;

	if( generalSoundError ) return;
	for(i=0; i<STREAM_COUNT; i++) {
		mmrAudioIdentifier[i] = NULL;
		mmrName[i] = NULL;
	}

	// open connections
	stream_open( STREAM_BGM, BGM_SOUTHERN );
	stream_open( STREAM_TOUCH_1, SFX_TOUCH_CLICK );
	stream_open( STREAM_TOUCH_2, SFX_TOUCH_CLICK );
	stream_open( STREAM_SHEEP_1, SFX_BAA_1 );
	stream_open( STREAM_SHEEP_2, SFX_BAA_2 );
	stream_open( STREAM_BEAR_GROWL, SFX_BEAR_GROWL );
	stream_open( STREAM_HORN, SFX_BEAR_HORN );
}

void destroy_audio() {
	if( generalSoundError ) return;
	stream_close( STREAM_BGM );
	stream_close( STREAM_TOUCH_1 );
	stream_close( STREAM_TOUCH_2 );
	stream_close( STREAM_SHEEP_1 );
	stream_close( STREAM_SHEEP_2 );
	stream_close( STREAM_BEAR_GROWL );
}

void stream_open( int which, int sound ) {
	int charCount;
	char currentDirectory[PATH_MAX];
	char resourceUrl[PATH_MAX];
	int audioOutputId;
	strm_dict_t *audioOutputParameters = NULL;
	mode_t fileAccessMode = S_IRUSR | S_IXUSR;

	if( generalSoundError ) return;

	getcwd(currentDirectory, PATH_MAX);
	charCount = snprintf(resourceUrl, PATH_MAX, "file://%s%s", currentDirectory, get_filename_for_sound( sound ) );
	if (charCount > PATH_MAX - 1) {
		generalSoundError = 1; return;
	}

	mmrAudioIdentifier[which] = mmrAudioIdentifierConstant;

	if ( ( mmrConnection[which] = mmr_connect( mmrName[which] ) ) == NULL ) {
		generalSoundError = 1; return;
	}

	if ( ( mmrContext[which] = mmr_context_create( mmrConnection[which], mmrContextName[which], 0, fileAccessMode ) ) == NULL ) {
		generalSoundError = 1; return;
	}

	if ( mmrAudioIdentifier[which] && ( audioOutputId = mmr_output_attach( mmrContext[which], mmrAudioIdentifier[which], "audio" ) ) < 0 ) {
		generalSoundError = 1; return;
	}

	if ( audioOutputParameters && mmr_output_parameters( mmrContext[which], audioOutputId, audioOutputParameters ) ) {
		generalSoundError = 1; return;
	}

	if ( mmr_input_attach( mmrContext[which], resourceUrl, mmrInputType ) < 0 ) {
		generalSoundError = 1; return;
	}
}

