/*
 * Copyright 2012 Sebastian Mihai
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * engine.c
 *
 *  Created on: 2012-08-10
 *      Author: Sebastian Mihai
 */

#include <math.h>
#include <stdlib.h>

#include "defs.h"
#include "graphics.h"
#include "engine.h"
#include "util.h"
#include "audio.h"

// determines how soon sheep run away from a bear
float DEGREE_OF_SHEEP_CAUTION = 2.0f;
// determines how soon bears run away from a dog
float DEGREE_OF_BEAR_CAUTION = 3.0f;

// an entity must be at least this much closer than another entity in order to be chosen
// by the closest-entity determination algorithm. The purpose of this is to remove "jitters"
// when between two entities from which an entity has to run. (such as a sheep between two bears!)
//
// a side effect of this is that, for example, a bear can be scarier only by virtue of existing
// first in the bears array, but this is not harmful in any way
float DEGREE_OF_CLOSEST_ENTITY_TOLERANCE = 100.0f;

#define SPAWNING_FRAME_COUNT 20

int scoreSheepSaved;
int scoreSheepEaten;

entity sheep[MAX_SHEEP];
entity bears[MAX_BEARS];
entity obstacles[MAX_OBSTACLES];

void game_statistics( int *saved, int *eaten ) {
	*saved = scoreSheepSaved;
	*eaten = scoreSheepEaten;
}

// return 1 if there's a collision, and 0 otherwise
int test_collision_by_entity( entity *a, entity *b ) {
	double dx, dy, d;
	dx = fabs(a->x - b->x);
	dy = fabs(a->y - b->y);
	d = sqrt( dx*dx + dy*dy );
	if( d <= (double)( max(a->radius, a->finalRadius) + max(b->radius, b->finalRadius)) ) {
		return 1;
	}
	return 0;
}

// same as test_collision_by_entity but specifying the second entity via its coords
int test_collision_by_coords( entity *a, float bx, float by, float bradius ) {
	double dx, dy, d;
	dx = fabs(a->x - bx);
	dy = fabs(a->y - by);
	d = sqrt( dx*dx + dy*dy );
	if( d <= (double)( max(a->radius, a->finalRadius) + bradius) ) {
		return 1;
	}
	return 0;
}

entity* find_nearest_entity( float x, float y, int type, double *distance, float tolerance ) {
	double dx, dy, d;
	entity *e = NULL;
	double minimum = 9999999.0;
	int i;

	if( type == ENTITY_BEAR ) {
		for( i=0; i<MAX_BEARS; i++ ) {
			if( bears[i].state != STATE_ON_SCREEN ) continue;
			dx = fabs(bears[i].x - x);
			dy = fabs(bears[i].y - y);
			d = sqrt( dx*dx + dy*dy );
			if( d < minimum - tolerance ) {
				minimum = d;
				e = &bears[i];
			}
		}
	} else if( type == ENTITY_OBSTACLE ) {
		for( i=0; i<MAX_OBSTACLES; i++ ) {
			if( obstacles[i].state != STATE_ON_SCREEN ) continue;
			dx = fabs(obstacles[i].x - x);
			dy = fabs(obstacles[i].y - y);
			d = sqrt( dx*dx + dy*dy );
			if( d < minimum - tolerance ) {
				minimum = d;
				e = &obstacles[i];
			}
		}
	} else if( type == ENTITY_SHEEP ) {
		for( i=0; i<MAX_SHEEP; i++ ) {
			if( sheep[i].state != STATE_ON_SCREEN ) continue;
			dx = fabs(sheep[i].x - x);
			dy = fabs(sheep[i].y - y);
			d = sqrt( dx*dx + dy*dy );
			if( d < minimum - tolerance ) {
				minimum = d;
				e = &sheep[i];
			}
		}
	}

	if( e ) {
		*distance = minimum;
	}

	return e;
}

int generate_spawn_point( entity *e ) {

	int i, found;
	int attempts = 0;
	float newx, newy;

	while( attempts < 100 ) {
		found = 1;
		attempts++;

		float boundaryBuffer = CONTACT_RADIUS+10.0f;

		// for obstacles, we'll leave a bit more room around the playfield boundaries
		if( e->type == ENTITY_OBSTACLE ) {
			boundaryBuffer = 4.0 * boundaryBuffer;
		}

		// at least CONTACT_RADIUS away from the edge of the screen
		newx = (float)boundedRandom(PLAYFIELD_WIDTH - 2*boundaryBuffer) - PLAYFIELD_WIDTH/2.0 + boundaryBuffer;
		newy = (float)boundedRandom(PLAYFIELD_HEIGHT - 2*boundaryBuffer) - PLAYFIELD_HEIGHT/2.0 + boundaryBuffer;

		// regardless of type, we don't want it on top of an obstacle
		for(i=0;i<MAX_OBSTACLES;i++) {
			if(obstacles[i].state == STATE_ON_SCREEN ) {

				float radiusCheck = max( e->radius, e->finalRadius );

				// for obstacles, we'll leave a bit more room in-between them
				if( e->type == ENTITY_OBSTACLE ) {
					radiusCheck = 4.0 * radiusCheck;
				}

				if( test_collision_by_coords(&obstacles[i], newx, newy, radiusCheck) ) {
					found = 0;
				}
			}
		}

		if( found ) {
			e->x = newx;
			e->y = newy;

			return EXIT_SUCCESS;
		}
	}

	return EXIT_FAILURE;
}

void spawn_bear() {
	int i;

	for(i=0;i<MAX_BEARS;i++) {
		if( bears[i].state == STATE_NOT_SPAWNED ) {
			bears[i].state = STATE_SPAWNING;
			bears[i].finalRadius = 50.0f + (float)boundedRandom( 100 ) / 10.0f;
			bears[i].textureIndex = assignTexture( ENTITY_BEAR );
			bears[i].angle = (float)boundedRandom( 628 );
			bears[i].speed = 1.00f + (float)boundedRandom( 850 ) / 1000.0f;
			bears[i].distance = 0.0f;
			bears[i].remainingFrames = SPAWNING_FRAME_COUNT;
			bears[i].radius = 0.0f;

			// if we can't find a spot for this bear, we don't spawn it
			if( generate_spawn_point( &bears[i] ) == EXIT_FAILURE ) {
				bears[i].state = STATE_NOT_SPAWNED;
				return;
			}
			stream_play( STREAM_BEAR_GROWL );
			return;
		}
	}
}

void spawn_sheep() {
	int i;

	for(i=0;i<MAX_SHEEP;i++) {
		if( sheep[i].state == STATE_NOT_SPAWNED ) {
			sheep[i].state = STATE_SPAWNING;
			sheep[i].finalRadius = 20.0f + (float)boundedRandom( 150 ) / 10.0f;

			if( boundedRandom (300) == 25 ) {
				sheep[i].finalRadius = 50.0f;
			}

			// 10 times the radians in a circle - eliminate some noise from the RNG
			sheep[i].angle = (float)boundedRandom( 628 );
			sheep[i].textureIndex = assignTexture( ENTITY_SHEEP );
			sheep[i].speed = 0.25f + (float)boundedRandom( 350 ) / 1000.0f;

			if( boundedRandom (300) == 25 ) {
				sheep[i].speed = 2.5f;
			}

			sheep[i].distance = 0.0f;
			sheep[i].remainingFrames = SPAWNING_FRAME_COUNT;
			sheep[i].radius = 0.0f;

			// if we can't find a spot for this sheep, we don't spawn it
			if( generate_spawn_point( &sheep[i] ) == EXIT_FAILURE ) {
				sheep[i].state = STATE_NOT_SPAWNED;
			}
			return;
		}
	}
}

void spawn_sheep_coords( int x, int y ) {
	int i;

	for(i=0;i<MAX_SHEEP;i++) {
		if( sheep[i].state == STATE_NOT_SPAWNED ) {
			sheep[i].state = STATE_SPAWNING;
			sheep[i].finalRadius = 20.0f + (float)boundedRandom( 150 ) / 10.0f;
			// 10 times the radians in a circle - eliminate some noise from the RNG
			sheep[i].angle = (float)boundedRandom( 628 );
			sheep[i].textureIndex = assignTexture( ENTITY_SHEEP );
			sheep[i].speed = 0.25f + (float)boundedRandom( 350 ) / 1000.0f;
			sheep[i].distance = 0.0f;
			sheep[i].remainingFrames = SPAWNING_FRAME_COUNT;
			sheep[i].radius = 0.0f;

			sheep[i].x = (float)x;
			sheep[i].y = (float)y;
			return;
		}
	}
}

void spawn_obstacle() {
	int i;

	for(i=0;i<MAX_OBSTACLES;i++) {
		if( obstacles[i].state == STATE_NOT_SPAWNED ) {
			obstacles[i].state = STATE_ON_SCREEN;
			obstacles[i].radius = 40.0f + (float)boundedRandom( 150 ) / 10.0f;
			obstacles[i].speed = 0.0f;
			obstacles[i].angle = (float)boundedRandom( 628 );
			obstacles[i].textureIndex = assignTexture( ENTITY_OBSTACLE );

			if( generate_spawn_point( &obstacles[i] ) == EXIT_FAILURE ) {
				obstacles[i].state = STATE_NOT_SPAWNED;
			}
			return;
		}
	}
}

void initialize_game() {
	int i;

	startRandomSequence();

	for(i=0;i<MAX_SHEEP;i++) {
		sheep[i].type = ENTITY_SHEEP;
		sheep[i].state = STATE_NOT_SPAWNED;
	}
	for(i=0;i<MAX_BEARS;i++) {
		bears[i].type = ENTITY_BEAR;
		bears[i].state = STATE_NOT_SPAWNED;
	}
	for(i=0;i<MAX_OBSTACLES;i++) {
		obstacles[i].type = ENTITY_OBSTACLE;
		obstacles[i].state = STATE_NOT_SPAWNED;
	}

	scoreSheepSaved = 1;
	scoreSheepEaten = 1;
}

void render_entities() {

	int i;

	for(i=0;i<MAX_OBSTACLES;i++) {
		if( obstacles[i].state == STATE_ON_SCREEN || obstacles[i].state == STATE_SPAWNING || obstacles[i].state == STATE_DESPAWNING ) {
			render_entity( &obstacles[i] );
		}
	}
	for(i=0;i<MAX_SHEEP;i++) {
		if( sheep[i].state == STATE_ON_SCREEN || sheep[i].state == STATE_SPAWNING || sheep[i].state == STATE_DESPAWNING ) {
			render_entity( &sheep[i] );
		}
	}
	for(i=0;i<MAX_BEARS;i++) {
		if( bears[i].state == STATE_ON_SCREEN || bears[i].state == STATE_SPAWNING || bears[i].state == STATE_DESPAWNING ) {
			render_entity( &bears[i] );
		}
	}
}

void sheep_act( int which ) {

	float newx, newy;

	if( sheep[which].state == STATE_ON_SCREEN ) {

		// do we have to run from any bears?
		double distance;
		entity *closestBear = find_nearest_entity( sheep[which].x, sheep[which].y, ENTITY_BEAR, &distance, DEGREE_OF_CLOSEST_ENTITY_TOLERANCE );

		if( closestBear ) {
			if( distance < DEGREE_OF_SHEEP_CAUTION * (sheep[which].radius + closestBear->radius ) ) {
				// this bear is close enough... run away from it!
				double dx = ( sheep[which].x - closestBear->x );
				double dy = ( sheep[which].y - closestBear->y );
				// calculate a new angle, leading sheep away from the bear!
				sheep[which].angle = (float)safe_atan2( dy, dx ) * 180.0 / PI;
			}
		}

		newx = sheep[which].x + cos(sheep[which].angle * PI / 180.0 ) * sheep[which].speed;
		newy = sheep[which].y + sin(sheep[which].angle * PI / 180.0 ) * sheep[which].speed;

		// collision with the playfield boundary
		if( newx < -PLAYFIELD_WIDTH/2.0 + sheep[which].radius ||
			newx > PLAYFIELD_WIDTH/2.0 - sheep[which].radius ||
			newy < -PLAYFIELD_HEIGHT/2.0 + sheep[which].radius ||
			newy > PLAYFIELD_HEIGHT/2.0 - sheep[which].radius ) {

			int sign = which % 2 == 0 ? -1 : 1;
			sheep[which].angle += 5.0f * (float)sign;
			return;
		}

		// did we collide with an obstacle?
		//entity *closestObstacle = find_nearest_entity( sheep[which].x, sheep[which].y, ENTITY_OBSTACLE, &distance, 0.0f );
		entity *closestObstacle = find_nearest_entity( newx, newy, ENTITY_OBSTACLE, &distance, 0.0f );
		if( closestObstacle ) {
			if( distance < sheep[which].radius + closestObstacle->radius ) {
				// collision... re-orient sheep, but don't move this frame

					//sheep[which].angle = (float)boundedRandom( 628 );
				int sign = which % 2 == 0 ? -1 : 1;
				sheep[which].angle += 5.0f * (float)sign;

				return;
			}
		}

		// move sheep

		sheep[which].distance += sheep[which].speed;
		sheep[which].x = newx;
		sheep[which].y = newy;

		if( sheep[which].distance >= 3.5f ) {
			sheep[which].textureIndex = nextSheepTexture( sheep[which].textureIndex );
			sheep[which].distance = 0.0f;
		}
	} else if( sheep[which].state == STATE_SPAWNING ) {
		if( sheep[which].remainingFrames == 0 ) {
			sheep[which].state = STATE_ON_SCREEN;
			return;
		}

		sheep[which].radius += sheep[which].finalRadius / SPAWNING_FRAME_COUNT;
		sheep[which].remainingFrames--;
	} else if( sheep[which].state == STATE_DESPAWNING ) {
		if( sheep[which].remainingFrames == 0 ) {
			sheep[which].state = STATE_NOT_SPAWNED;
			return;
		}

		sheep[which].radius -= sheep[which].finalRadius / SPAWNING_FRAME_COUNT;
		sheep[which].angle += (SPAWNING_FRAME_COUNT / 2.0 - (float)sheep[which].remainingFrames) * ( 360.0f / (SPAWNING_FRAME_COUNT / 2.0f ) );
		sheep[which].remainingFrames--;
	}
}

void bear_act( int which ) {
	float newx, newy;

	if( bears[which].state == STATE_ON_SCREEN ) {

		double distance;

		newx = bears[which].x + cos(bears[which].angle * PI / 180.0 ) * bears[which].speed;
		newy = bears[which].y + sin(bears[which].angle * PI / 180.0 ) * bears[which].speed;

		// collision with the playfield boundary
		if( newx < -PLAYFIELD_WIDTH/2.0 + bears[which].radius ||
			newx > PLAYFIELD_WIDTH/2.0 - bears[which].radius ||
			newy < -PLAYFIELD_HEIGHT/2.0 + bears[which].radius ||
			newy > PLAYFIELD_HEIGHT/2.0 - bears[which].radius ) {

			int sign = which % 2 == 0 ? -1 : 1;
			bears[which].angle += 5.0f * (float)sign;
			return;
		}

		float firstMin = ( bears[which].x - (-PLAYFIELD_WIDTH/2.0) );
		float secondMin = ( (PLAYFIELD_WIDTH/2.0) - bears[which].x );
		float thirdMin = ( bears[which].y - (-PLAYFIELD_HEIGHT/2.0) );
		float fourthMin = ( (PLAYFIELD_HEIGHT/2.0) - bears[which].y );

		double distanceToPlayfieldBoundary = min( min( firstMin, secondMin), min( thirdMin, fourthMin ) );

		// did we collide with an obstacle?
		entity *closestObstacle = find_nearest_entity( newx, newy, ENTITY_OBSTACLE, &distance, 0.0f );
		if( closestObstacle ) {
			if( distance < bears[which].radius + closestObstacle->radius ) {
				// collision... re-orient sheep, but don't move this frame

					//bears[which].angle = (float)boundedRandom( 628 );
				int sign = which % 2 == 0 ? -1 : 1;
				bears[which].angle += 5.0f * (float)sign;

				return;
			}
		}

		// move bear

		bears[which].distance += bears[which].speed;
		bears[which].x = newx;
		bears[which].y = newy;

		if( bears[which].distance >= 6.0f ) {
			bears[which].textureIndex = nextBearTexture( bears[which].textureIndex );
			bears[which].distance = 0.0f;
		}

		int nearBoundary = 0;
		int nearObstacle = 0;

		if( closestObstacle ) {
			if( distance < (bears[which].radius + closestObstacle->radius) + 30.0f ) {
				// close to an obstacle - don't chase a sheep
				nearObstacle = 1;
			}
		}

		if( distanceToPlayfieldBoundary < (bears[which].radius) + 30.0f ) {
			// close to the boundary - don't change angle or chase sheep
			nearBoundary = 1;
		}

		if( nearBoundary || nearObstacle ) {
			// small chance to move towards middle of screen
			if( nearBoundary && boundedRandom( 100 ) == 25 ) {
				newx = (float)boundedRandom( 400 ) - 200.0f;
				newy = (float)boundedRandom( 400 ) - 200.0f;
				double dx = ( newx - bears[which].x );
				double dy = ( newy - bears[which].y );
				// calculate a new angle, leading towards the sheep
				bears[which].angle = (float)safe_atan2( dy, dx ) * 180.0 / PI;
			}
			return;
		}

		// find a sheep to chase
		entity *closestSheep = find_nearest_entity( bears[which].x, bears[which].y, ENTITY_SHEEP, &distance, 0.0f );

		if( closestSheep && distance < 300.0f ) {
			// this sheep is close enough... catch it!
			double dx = ( closestSheep->x - bears[which].x );
			double dy = ( closestSheep->y - bears[which].y );
			// calculate a new angle, leading towards the sheep
			bears[which].angle = (float)safe_atan2( dy, dx ) * 180.0 / PI;
		}
	} else if( bears[which].state == STATE_SPAWNING ) {
		if( bears[which].remainingFrames == 0 ) {
			bears[which].state = STATE_ON_SCREEN;
			return;
		}

		bears[which].radius += bears[which].finalRadius / SPAWNING_FRAME_COUNT;
		bears[which].remainingFrames--;
	} else if( bears[which].state == STATE_DESPAWNING ) {
		if( bears[which].remainingFrames == 0 ) {
			bears[which].state = STATE_NOT_SPAWNED;
			return;
		}

		bears[which].radius -= bears[which].finalRadius / SPAWNING_FRAME_COUNT;
		bears[which].angle += (SPAWNING_FRAME_COUNT / 2.0 - (float)bears[which].remainingFrames) * ( 360.0f / (SPAWNING_FRAME_COUNT / 2.0f ) );
		bears[which].remainingFrames--;
	}
}

void entities_act() {
	int i;

	for(i=0;i<MAX_SHEEP;i++) {
		sheep_act( i );
	}

	for(i=0;i<MAX_BEARS;i++) {
		bear_act( i );
	}
}

void sheep_collision( int which ) {

	if( sheep[which].state == STATE_ON_SCREEN ) {
		double distance;
		entity *closestBear = find_nearest_entity( sheep[which].x, sheep[which].y, ENTITY_BEAR, &distance, 0.0f );

		if( closestBear ) {
			// have the entities overlap just a bit, so that it looks more realistic
			if( distance < sheep[which].radius + closestBear->radius - 10.0f ) {
				// this bear has caught the sheep...
				sheep[which].state = STATE_DESPAWNING;
				sheep[which].remainingFrames = SPAWNING_FRAME_COUNT;
				scoreSheepEaten++;

				if( boundedRandom(2) == 0 ) {
					stream_play( STREAM_SHEEP_1 );
				} else {
					stream_play( STREAM_SHEEP_2 );
				}
			}
		}
	}
}

void sheep_saved( int which, float x, float y, float radius ) {

	if( sheep[which].state == STATE_ON_SCREEN || sheep[which].state == STATE_SPAWNING ) {
		if( test_collision_by_coords( &sheep[which], x, y, radius ) ) {
			sheep[which].state = STATE_DESPAWNING;
			sheep[which].remainingFrames = SPAWNING_FRAME_COUNT / 2;
			scoreSheepSaved++;
		}
	}
}

void detect_collisions() {
	int i;

	for(i=0;i<MAX_SHEEP;i++) {
		sheep_collision( i );
	}
}

void detect_safe_sheep( float x, float y, float radius ) {
	int i;

	for(i=0;i<MAX_SHEEP;i++) {
		sheep_saved( i, x, y, radius );
	}
}

void bear_scared( int which, float x, float y, float radius ) {

	if( bears[which].state == STATE_ON_SCREEN || bears[which].state == STATE_SPAWNING ) {
		if( test_collision_by_coords( &bears[which], x, y, radius ) ) {
			bears[which].state = STATE_DESPAWNING;
			bears[which].remainingFrames = SPAWNING_FRAME_COUNT / 2;
		}
	}
}

void detect_scared_bear( float x, float y, float radius ) {
	int i;

	for(i=0;i<MAX_BEARS;i++) {
		bear_scared( i, x, y, radius );
	}
}
