/*
 * Copyright 2012 Sebastian Mihai
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * input.c
 *
 *  Created on: 2012-08-13
 *      Author: Sebastian Mihai
 */

#include <bps/screen.h>
#include <stdlib.h>

#include "defs.h"

int surface_width, surface_height;

static int globalViewportX = 0;
static int globalViewportY = 0;
static int viewportX = 0;
static int viewportY = 0;

// consumers of this module will supply a callback for scrolling
static void (*scrollHandler)(float, float);

// consumers of this module will supply a callback for handling single touch events
static void (*singleTouchHandler)(float, float);

void handleScroll( int x, int y ) {

	int leftScrollBound = -PLAYFIELD_WIDTH/2 + surface_width/2 - PLAYFIELD_PADDING;
	int rightScrollBound = -leftScrollBound;
	int bottomScrollBound = -PLAYFIELD_HEIGHT/2 + surface_height/2 - PLAYFIELD_PADDING;
	int topScrollBound = -bottomScrollBound;

	if( scrollHandler == NULL ) {
		return;
	}

	// if the scroll amount is more than half the size of the screen, it's probably a result of
	// the user touching the screen in two places at once, so we ignore it
	if( abs( x - viewportX ) > surface_width/2 ||
		abs( viewportY - y ) > surface_height/2 ) {
		return;
	}

	// enforce scroll boundaries
	if( globalViewportX + viewportX - x < leftScrollBound ) {
		x -= leftScrollBound - (viewportX - x) - globalViewportX;
	}

	if( globalViewportX + viewportX - x > rightScrollBound ) {
		x += globalViewportX + (viewportX - x) - rightScrollBound;
	}

	if( globalViewportY + y - viewportY < bottomScrollBound ) {
		y += bottomScrollBound - (y - viewportY) - globalViewportY;
	}

	if( globalViewportY + y - viewportY > topScrollBound ) {
		y -= globalViewportY + (y - viewportY) - topScrollBound;
	}

	scrollHandler( (float)(x-viewportX),
			 	   (float)(viewportY-y) );

	globalViewportX += viewportX - x;
	globalViewportY += y - viewportY;
	viewportX = x;
	viewportY = y;
}

void handleTouch( int x, int y ) {
	viewportX = x;
	viewportY = y;
}

void handleSingleTouchAndRelease( int x, int y ) {

	if( singleTouchHandler == NULL ) {
		return;
	}

	singleTouchHandler( globalViewportX + (x - surface_width/2),
					    (surface_height/2 - y) + globalViewportY );

	viewportX = 0;
	viewportY = 0;
}

void handleEndOfScroll() {
	viewportX = 0;
	viewportY = 0;
}

void handleScreenEvent(bps_event_t *event) {

	// store initial (entry) touch
	static int initialTouchCoords[2] = {-9999, -9999};

	static int lastTouchCoords[2];
	static short int touchHeld = 0;
	int currentPositionCoords[2];

    screen_event_t screen_event = screen_event_get_event(event);

    int screen_val;
    screen_get_event_property_iv(screen_event, SCREEN_PROPERTY_TYPE, &screen_val);

    screen_get_event_property_iv(screen_event, SCREEN_PROPERTY_POSITION, currentPositionCoords);

    switch (screen_val) {
    case SCREEN_EVENT_MTOUCH_TOUCH:
    	if( !touchHeld ) {
    		// save where we started the touch
    		screen_get_event_property_iv(screen_event, SCREEN_PROPERTY_SOURCE_POSITION,
    		                lastTouchCoords);

    		initialTouchCoords[0] = currentPositionCoords[0];
    		initialTouchCoords[1] = currentPositionCoords[1];
    		lastTouchCoords[0] = currentPositionCoords[0];
    		lastTouchCoords[1] = currentPositionCoords[1];

    		handleTouch( initialTouchCoords[0], initialTouchCoords[1] );
    		touchHeld = 1;
    	}
    	break;
    case SCREEN_EVENT_MTOUCH_MOVE:
    	// de-sensitize scrolling somewhat, so that users don't inadvertently scroll when they wanted to touch
    	if( touchHeld > 3) {
    		/*glTranslatef((float)(currentPositionCoords[0]-lastTouchCoords[0]),
    					 (float)(lastTouchCoords[1]-currentPositionCoords[1]),
    					 0.0f);*/
    		handleScroll( currentPositionCoords[0], currentPositionCoords[1] );
    	}

		lastTouchCoords[0] = currentPositionCoords[0];
		lastTouchCoords[1] = currentPositionCoords[1];
		touchHeld++;
		break;
    case SCREEN_EVENT_MTOUCH_RELEASE:

    	// if there was just a MINOR movement, we treat it as a single touch
    	if( touchHeld <= 3) {
    		handleSingleTouchAndRelease( initialTouchCoords[0], initialTouchCoords[1] );
    	} else {
    		handleEndOfScroll();
    	}

    	touchHeld = 0;

    	initialTouchCoords[0] = -9999;
    	initialTouchCoords[1] = -9999;
        break;
    }
}


void initialize_input( int surfaceWidth, int surfaceHeight ) {
	globalViewportX = 0;
	globalViewportY = 0;
	viewportX = 0;
	viewportY = 0;
	surface_width = surfaceWidth;
	surface_height = surfaceHeight;
}

void addScrollHandler( void (*scrlHndlr)(float, float) ) {
	scrollHandler = scrlHndlr;
}

void addSingleTouchHandler( void (*snglTchHndlr)(float, float) ) {
	singleTouchHandler = snglTchHndlr;
}
