/*
 * Copyright 2012 Sebastian Mihai
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * manager.c
 *
 *  Created on: 2012-08-13
 *      Author: Sebastian Mihai
 */

#include <bps/navigator.h>
#include <bps/screen.h>
#include <stdlib.h>
#include <assert.h>

#include "bbutil.h"
#include "defs.h"
#include "audio.h"
#include "graphics.h"
#include "engine.h"
#include "input.h"
#include "util.h"

#define TOUCH_RADIUS 100.0f
#define TOUCH_FRAMES 7

#define POWERUP_NONE 0
#define POWERUP_HORN 1

#define MANAGER_STATE_MENU 0
#define MANAGER_STATE_GAME 1

#define GAME_OVER_AFTER_EATEN 50
#define POWERUP_AFTER_EVERY 30

int saveFrames = 0;
float saveX, saveY;

int lastPowerupCount;
int gameOverFrames;
int menuTaps = 0;
int powerupCharges = 0;
int powerup = POWERUP_NONE;
int currentState = MANAGER_STATE_MENU;

int sheepSpawnChanceOutOf = 30;
int bearSpawnChanceOutOf = 150;
unsigned long currentFrame = 0;
int sheepSavedCount = -1;
int sheepEatenCount = -1;

void handleSingleTouchEvent ( float x, float y );
void menu_handleSingleTouchEvent ( float x, float y );

void beginGame() {
	int surface_width, surface_height;

	// add a scroll handler to the input module, defined in the graphics module
	addScrollHandler( scroll_screen );
	// add a single touch handler to the input module
	addSingleTouchHandler( handleSingleTouchEvent );

	initialize_graphics( &surface_width, &surface_height );
	initialize_input( surface_width, surface_height );

	initialize_game();

	int i;
	for(i=0;i<MAX_OBSTACLES;i++)spawn_obstacle();

	powerupCharges = 0;
	powerup = POWERUP_NONE;
	currentFrame = 0;
	saveFrames = 0;
	sheepSavedCount = 1;
	sheepEatenCount = 1;
	gameOverFrames = 100;
	lastPowerupCount = 0;
}

void beginMenu() {
	int surface_width, surface_height;

	// no scroll handler - we do not scroll while in menu mode
	addScrollHandler( NULL );
	// our menu single touch handler
	addSingleTouchHandler( menu_handleSingleTouchEvent );

	initialize_graphics( &surface_width, &surface_height );
	initialize_input( surface_width, surface_height );

	// if the game ended holding powerup, reset touch stream to clicking
	stream_close( STREAM_TOUCH_1 );
	stream_open( STREAM_TOUCH_1, SFX_TOUCH_CLICK );
	stream_play( STREAM_TOUCH_1 );

	currentFrame = 0;
	menuTaps = 0;
}

/*
 * Single touch handler for the menu
 */
void menu_handleSingleTouchEvent ( float x, float y ) {
	menuTaps++;
	stream_play( STREAM_TOUCH_1 );
	if( menuTaps == 2 ) {
		currentState = MANAGER_STATE_GAME;
		beginGame();
	}
}

/*
 * Single touch handler for the game
 */
void handleSingleTouchEvent ( float x, float y ) {

	// alternate between the two "click" streams so that the clicking is more responsive
	static int clickStream = STREAM_TOUCH_1;

	if( powerup == POWERUP_NONE ) {
		stream_play( clickStream );
		clickStream = clickStream == STREAM_TOUCH_1 ? STREAM_TOUCH_2 : STREAM_TOUCH_1;
		detect_safe_sheep( x, y, (float)TOUCH_RADIUS );
	} else if( powerup == POWERUP_HORN ) {
		stream_play( STREAM_HORN );
		detect_scared_bear( x, y, (float)TOUCH_RADIUS );
		powerupCharges--;
		if( powerupCharges == 0 ) {
			powerup = POWERUP_NONE;
		}
	}
	saveFrames = TOUCH_FRAMES;
	saveX = x;
	saveY = y;
}

int initialize_game_manager() {

	load_textures();

    initialize_audio();

    // HACK... the first time these streams are played, they sound NOTHING like what the
    // sound effect should be, so just play them here for no reason :)
    stream_play( STREAM_TOUCH_1 );
    stream_play( STREAM_TOUCH_2 );

    currentState = MANAGER_STATE_MENU;
    beginMenu();

    return EXIT_SUCCESS;
}

void run_frame_game() {

	if( sheepEatenCount >= GAME_OVER_AFTER_EATEN ) {
		// GAME OVER!!!
		gameOverFrames--;
		stream_play( STREAM_SHEEP_2 );

		render_begin();
		render_playfield();
		render_entities();
		render_end();
		render_swap();

		if( gameOverFrames == 0 ) {
			currentState = MANAGER_STATE_MENU;
			beginMenu();
		}
		return;
	}

	// sometimes spawn sheep
	if( boundedRandom( sheepSpawnChanceOutOf ) == 0 ) {
		spawn_sheep();
	}

	// sometimes spawn bears
	if( boundedRandom( bearSpawnChanceOutOf ) == 0 ) {
		spawn_bear();
	}

	entities_act();
	detect_collisions();

	render_begin();
	render_playfield();
	render_entities();

	// if there was a touch, show the area of effect
	if( saveFrames > 0 ) {
		saveFrames--;
		render_save_area_of_effect( saveX, saveY, TOUCH_RADIUS * (float)(TOUCH_FRAMES-saveFrames) / (float)TOUCH_FRAMES );
	}

	if( powerup == POWERUP_HORN ) {
		if( currentFrame % 30 < 15 ) {
			render_powerup();
		}
	} else {
		render_powerup_partial( ((float)sheepSavedCount - (float)lastPowerupCount) / (float)POWERUP_AFTER_EVERY );
	}

	render_end();
	render_swap();

	game_statistics( &sheepSavedCount, &sheepEatenCount );

	// did we get a powerup?
	if( sheepSavedCount - lastPowerupCount >= POWERUP_AFTER_EVERY ) {
		stream_play( STREAM_HORN );
		lastPowerupCount = sheepSavedCount;
		powerupCharges = 5;
		powerup = POWERUP_HORN;
	}
}

void run_frame_menu() {

	int score;

	render_begin();
	render_playfield();
	render_menu();

	// calculate score
	if( sheepSavedCount != -1 ) {

		if( sheepSavedCount < 45 ) {
			score = 0;
		} else if( sheepSavedCount < 75 ) {
			score = 1;
		} else if( sheepSavedCount < 100 ) {
			score = 2;
		} else if( sheepSavedCount < 125 ) {
			score = 3;
		} else if( sheepSavedCount < 155 ) {
			score = 4;
		} else {
			score = 5;
		}

		render_score( score );
	}


	render_end();
	render_swap();
}

void manager_run_frame() {

	currentFrame++;
	// introduce a bit of chaos
	boundedRandom( 10 );

	stream_play( STREAM_BGM );

	if( currentState == MANAGER_STATE_GAME ) {
		run_frame_game();
	} else {
		run_frame_menu();
	}
}

void manager_terminate() {
	destroy_audio();
}

void manager_handle_screen_event( bps_event_t *event ) {
	// delegate to the input module
	handleScreenEvent( event );
}
